library jaguar_orm.adapter;

import 'dart:async';
import 'package:jaguar_query/jaguar_query.dart';

/// Adapter interface that must be implemented to support new databases
abstract class Adapter<ConnType> {
  /// Makes a new connection to database
  Future<Null> connect();

  /// Closes the connection
  Future<Null> close();

  ConnType get connection;

  /// Returns a row found by executing [statement]
  Future<Map> findOne(Find statement);

  /// Returns a list of rows found by executing [statement]
  Future<Stream<Map>> find(Find statement);

  /// Executes the insert statement and returns the primary key of
  /// inserted row
  Future<dynamic> insert(Insert statement);

  /// Updates the row and returns the number of rows updated
  Future<int> update(Update statement);

  /// Deletes the requested row
  Future<int> delete(Delete statement);

  /// Creates the table
  Future<Null> createTable(Create statement);

  /// Create the database
  Future<Null> createDatabase(CreateDb statement);

  /// Drops tables from database
  Future<Null> dropTable(Drop st);
}

abstract class Mapper<ModelType> {
  ModelType fromMap(Map map);
}

class BuiltFind<ConnType, T> {
  final Adapter<ConnType> adapter;

  final Mapper<T> mapper;

  //TODO make it unmodifiable
  final Find st;

  BuiltFind(this.adapter, this.st, [this.mapper]);

  /// Returns a row found by executing [statement]
  Future<Map> one() => adapter.findOne(st);

  /// Returns a row found by executing [statement]
  Future<T> oneMapped() async {
    final Map map = await adapter.findOne(st);
    return mapper.fromMap(map);
  }

  /// Returns a row found by executing [statement]
  Future<List<Map>> many() async => (await adapter.find(st)).toList();

  /// Returns a row found by executing [statement]
  Future<Stream<Map>> manyStream() => adapter.find(st);

  /// Returns a row found by executing [statement]
  Future<List<T>> manyMapped() async =>
      (await adapter.find(st)).map(mapper.fromMap).toList();

  /// Returns a row found by executing [statement]
  Future<Stream<T>> manyStreamMapped() async =>
      (await adapter.find(st)).map(mapper.fromMap);
}
